# Copyright 2020, 2021 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0
"""
The Lissp Read-Evaluate-Print Loop. For interactive use.
"""

import sys
from code import InteractiveConsole
from types import ModuleType, SimpleNamespace

import hissp.basic
from hissp.reader import Lissp, SoftSyntaxError


class LisspREPL(InteractiveConsole):
    """Lissp's Read-Evaluate-Print Loop, layered on Python's.

    You can initialize the REPL with a locals dict,
    which is useful for debugging other modules.
    Call interact() to start.
    """

    def __init__(self, locals=None, filename="<console>"):
        super().__init__(locals, filename)
        sys.ps1 = "#> "
        sys.ps2 = "#.."
        self.lissp = Lissp(ns=locals)
        self.locals = self.lissp.ns

    def runsource(self, source, filename="<input>", symbol="single"):
        """:meta private:"""
        try:
            self.lissp.filename = filename
            source = self.lissp.compile(source)
        except SoftSyntaxError:
            return True
        except SyntaxError:
            self.showsyntaxerror()
            return False
        except BaseException:
            import traceback

            traceback.print_exc()
            return False
        print(">>>", source.replace("\n", "\n... "), file=sys.stderr)
        super().runsource(source, filename, symbol)


def force_main():
    """:meta private:"""
    __main__ = ModuleType("__main__")
    sys.modules["__main__"] = __main__
    sys.path.insert(0, "")
    return __main__


def main(__main__=None):
    """REPL command-line entry point."""
    if not __main__:
        __main__ = force_main()
    repl = LisspREPL(locals=__main__.__dict__)
    repl.locals["_macro_"] = SimpleNamespace(**vars(hissp.basic._macro_))
    repl.interact()


if __name__ == "__main__":
    main()



# Copyright 2020 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0
"""
Hissp's command-line interface.
"""

import argparse
import re
import sys

import hissp.repl
from hissp.reader import Lissp


def main():
    """
    Entry point for the `lissp command`.
    """
    args = _arg_parser().parse_args()
    sys.argv = [""]
    __main__ = hissp.repl.force_main()
    if args.c is not None:
        _cmd(args, __main__.__dict__)
    elif args.file is not None:
        _with_args(args, __main__.__dict__)
    else:
        hissp.repl.main(__main__)


def _cmd(args, ns):
    sys.argv = ["-c"]
    if args.file is not None:
        sys.argv.extend([args.file, *args.args])
    args.i("(hissp.basic.._macro_.prelude)\n" + args.c, ns)


def _with_args(args, ns):
    with argparse.FileType("r")(args.file) as file:
        sys.argv = [file.name, *args.args]
        code = file.read()
    args.i(re.sub("^#!.*\n", "\n", code), ns)


def _interact(code, ns):
    repl = hissp.repl.LisspREPL(locals=ns)
    repl.lissp.compiler.evaluate = True
    try:
        repl.lissp.compile(code)
    finally:
        repl.lissp.compiler.evaluate = False
        repl.interact()


def _no_interact(code, ns):
    Lissp(ns=ns, evaluate=True).compile(code)


def _arg_parser():
    root = argparse.ArgumentParser(
        description="Starts the REPL if there are no arguments."
    )
    _ = root.add_argument
    _(
        "-i",
        action="store_const",
        const=_interact,
        default=_no_interact,
        help="Drop into REPL after the script.",
    )
    _("-c", help="Run main script (with prelude) from this string.", metavar="cmd")
    _("file", nargs="?", help="Run main script from this file. (- for stdin.)")
    _("args", nargs="*", help="Arguments for the script.")
    return root


if __name__ == "__main__":
    main()



# Copyright 2019, 2020 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0
"""
The Lissp language reader and associated helper functions.

Compiles Lissp files to Python files when run as the main module.

The reader is organized as a lexer and parser.
The parser is extensible with Lissp reader macros.
The lexer is not extensible,
and doesn't do much more than pull tokens from a relatively simple regex
and track its position for error messages.
"""

import ast
import builtins
import os
import re
import sys
from contextlib import contextmanager, nullcontext
from functools import reduce
from importlib import import_module, resources
from itertools import chain
from pathlib import Path, PurePath
from pprint import pformat, pprint
from types import ModuleType
from typing import Any, Iterable, Iterator, NewType, Optional, Tuple, Union

from hissp.compiler import Compiler, readerless
from hissp.munger import munge

# fmt: off
ENTUPLE = ("lambda",(":",":*","xAUTO0_"),"xAUTO0_",)
# fmt: on
"""
Used by the template macro to make tuples.

To avoid creating a dependency on Hissp, by default,
templates spell out the entuple implementation every time,
but you can override this by setting some other value here.
"""

TOKENS = re.compile(
    r"""(?x)
 (?P<comment>;.*)
|(?P<whitespace>[\n ]+)
|(?P<badspace>\s)  # Other whitespace not allowed.
|(?P<open>\()
|(?P<close>\))
|(?P<macro>
   ,@
  |['`,]
   # Any atom that ends in (an unescaped) ``#``
  |(?:[^\\ \n"();#]|\\.)+[#]
 )
|(?P<string>
  [#]?  # raw?
  "  # Open quote.
    (?:[^"\\]  # Any non-magic character.
       |\\(?:.|\n)  # Backslash only if paired, including with newline.
    )*  # Zero or more times.
  "  # Close quote.
 )
|(?P<continue>")
|(?P<atom>(?:[^\\ \n"();]|\\.)+)  # Let Python deal with it.
|(?P<error>.)
"""
)

Token = NewType("Token", Tuple[str, str, int])

DROP = object()
"""
The sentinel value returned by the discard macro ``_#``, which the 
reader skips over when parsing. Reader macros can have read-time side 
effects with no Hissp output by returning this.
"""


class SoftSyntaxError(SyntaxError):
    """A syntax error that could be corrected with more lines of input.

    When the REPL encounters this when attempting to evaluate a form,
    it will ask for more lines, rather than aborting with an error.
    """


class Lexer(Iterator):
    """The tokenizer for the Lissp language.

    Most of the actual tokenizing is done by the regex.
    The Lexer adds some position tracking to that to help with error
    messages.
    """

    def __init__(self, code: str, file: str = "<?>"):
        self.code = code
        self.file = file
        self.it = self._it()

    def __iter__(self) -> Iterator[Token]:
        return self

    def __next__(self):
        return next(self.it)

    def _it(self):
        pos = 0
        while pos < len(self.code):
            match = TOKENS.match(self.code, pos)
            assert match is not None
            assert match.lastgroup
            assert match.end() > pos, match.groups()
            pos = match.end()
            yield Token((match.lastgroup, match.group(), pos))

    def position(self, pos: int) -> Tuple[str, int, int, str]:
        """
        Compute the ``filename``, ``lineno``, ``offset`` and ``text``
        for a `SyntaxError`, using the current character offset in code.
        """
        good = self.code[0:pos].split("\n")
        lineno = len(good)
        offset = len(good[-1])
        return self.file, lineno, offset, self.code.split("\n")[lineno - 1]


class _Unquote(tuple):
    def __repr__(self):
        return f"_Unquote{super().__repr__()}"


def gensym_counter(count=[0]):
    """
    Call to increment the gensym counter, and return the new count.
    Used by the gensym reader macro ``$#`` to ensure symbols are unique.
    """
    count[0] += 1
    return count[0]


class Lissp:
    """
    The parser for the Lissp language.

    Wraps around a Hissp compiler instance.
    Parses Lissp tokens into Hissp syntax trees.
    """

    def __init__(
        self,
        qualname="__main__",
        ns=None,
        verbose=False,
        evaluate=False,
        filename="<?>",
    ):
        self.qualname = qualname
        self.compiler = Compiler(self.qualname, ns, evaluate)
        self.verbose = verbose
        self.filename = filename
        self.reinit()

    @property
    def ns(self):
        """The wrapped `Compiler`'s ``ns``."""
        return self.compiler.ns

    @ns.setter
    def ns(self, ns):
        self.compiler.ns = ns

    def reinit(self):
        """Reset position, nesting depth, and gensym stack."""
        self.gensym_stack = []
        self.depth = 0
        self._p = 0

    def position(self):
        """
        Get the ``filename``, ``lineno``, ``offset`` and ``text``
        for a `SyntaxError`, from the `Lexer` given to `parse`.
        """
        return self.tokens.position(self._p)

    def parse(self, tokens: Lexer) -> Iterator:
        """Build Hissp forms from a `Lexer`."""
        self.tokens = tokens
        return (form for form in self._parse() if form is not DROP)

    def _parse(self) -> Iterator:
        for k, v, self._p in self.tokens:
            if k in {"comment", "whitespace"}:
                continue
            elif k == "badspace":
                raise SyntaxError(
                    repr(v) + " is not whitespace in Lissp. Indent with spaces only.",
                    self.position(),
                )
            elif k == "open":
                yield from self._open()
            elif k == "close":
                self._close()
                return
            elif k == "macro":
                yield from self._macro(v)
            elif k == "string":
                yield self._string(v)
            elif k == "continue":
                raise SoftSyntaxError("Incomplete token.", self.position())
            elif k == "atom":
                yield self._atom(v)
            elif k == "error":
                raise SyntaxError("Can't read this.", self.position())
            else:
                assert False, "unknown token: " + repr(k)
        if self.depth:
            raise SoftSyntaxError(
                "Ran out of tokens before completing form.", self.position()
            )

    def _open(self):
        depth = self.depth
        self.depth += 1
        yield (*self.parse(self.tokens),)
        if self.depth != depth:
            raise SoftSyntaxError("Unclosed '('.", self.position())

    def _close(self):
        self.depth -= 1
        if self.depth < 0:
            raise SyntaxError("Unopened ')'.", self.position())

    @staticmethod
    def _string(v):
        if v[0] == "#":  # Let Python process escapes.
            v = v.replace("\\\n", "").replace("\n", r"\n")
            val = ast.literal_eval(v[1:])
        else:  # raw
            val = v[1:-1]  # Only remove quotes.
        return v if (v := pformat(val)).startswith("(") else f"({v})"

    def _macro(self, v):
        with {
            "`": self.gensym_context,
            ",": self.unquote_context,
            ",@": self.unquote_context,
        }.get(v, nullcontext)():
            try:
                depth = self.depth
                form = next(self.parse(self.tokens))
            except StopIteration:
                if self.depth == depth:
                    raise SoftSyntaxError(
                        f"Reader macro {v!r} missing argument.", self.position()
                    ) from None
                raise SyntaxError(
                    f"Reader macro {v!r} missing argument.", self.position()
                ) from None
            yield self.parse_macro(v, form)

    @staticmethod
    def _atom(v):
        is_symbol = "\\" == v[0]
        v = Lissp.escape(v)
        if is_symbol:
            return munge(v)
        try:
            val = ast.literal_eval(v)
            if isinstance(val, bytes):  # bytes have their own literals.
                return munge(v)
            return val
        except (ValueError, SyntaxError):
            return munge(v)

    def parse_macro(self, tag: str, form):
        """Apply a reader macro to a form."""
        if tag == "'":
            return "quote", form
        if tag == "`":
            return self.template(form)
        if tag == ",":
            return _Unquote([":?", form])
        if tag == ",@":
            return _Unquote([":*", form])
        assert tag.endswith("#")
        return self._parse_tag(tag[:-1], form)

    def _parse_tag(self, tag, form):
        if tag == "_":
            return DROP
        if tag == "$":
            return self.gensym(form)
        if tag == ".":
            return eval(readerless(form, self.ns), self.ns)
        if is_string(form):
            form = ast.literal_eval(form)
        tag = munge(self.escape(tag))
        if ".." in tag and not tag.startswith(".."):
            module, function = tag.split("..", 1)
            return reduce(getattr, function.split("."), import_module(module))(form)
        try:
            m = getattr(self.ns["_macro_"], tag + "xHASH_")
        except (AttributeError, KeyError):
            raise SyntaxError(f"Unknown reader macro {tag}", self.position())
        with self.compiler.macro_context():
            return m(form)

    @staticmethod
    def escape(atom):
        """Process the backslashes in a token."""
        atom = atom.replace(r"\.", "xFULLxSTOP_")
        return re.sub(r"\\(.)", lambda m: m[1], atom)

    def template(self, form):
        """Process form as template."""
        case = type(form)
        if is_string(form):
            return "quote", form
        if case is tuple and form:
            return (
                ENTUPLE,
                ":",
                *chain(*self._template(form)),
            )
        if case is str and not form.startswith(":"):
            return "quote", self.qualify(form)
        if case is _Unquote and form[0] == ":?":
            return form[1]
        return form

    def _template(self, forms: Iterable) -> Iterable[Tuple[str, Any]]:
        invocation = True
        for form in forms:
            case = type(form)
            if case is str and not form.startswith(":"):
                yield ":?", ("quote", self.qualify(form, invocation))
            elif case is _Unquote:
                yield form
            elif case is tuple:
                yield ":?", self.template(form)
            else:
                yield ":?", form
            invocation = False

    def qualify(self, symbol: str, invocation=False) -> str:
        """Qualify symbol based on current context."""
        if re.search(
            r"^\(|^\.|\.$|^quote$|^lambda$|^__import__$|xAUTO\d+_$|\.\.", symbol
        ):
            return symbol  # Not qualifiable.
        if invocation and "_macro_" in self.ns and self._macro_has(symbol):
            return f"{self.qualname}.._macro_.{symbol}"  # Known macro.
        if symbol in dir(builtins) and symbol.split(".", 1)[0] not in self.ns:
            return f"builtins..{symbol}"  # Known builtin, not shadowed (yet).
        if invocation and "." not in symbol:  # Could still be a recursive macro.
            return f"{self.qualname}..xAUTO_.{symbol}"
        return f"{self.qualname}..{symbol}"

    def _macro_has(self, symbol):
        # The _macro_ interface is not required to implement
        # __contains__ or __dir__ and exotic _macro_ objects might
        # override __getattribute__. The only way to tell if _macro_ has
        # a name is getattr().
        try:
            getattr(self.ns["_macro_"], symbol)
        except AttributeError:
            return False
        return True

    def reads(self, code: str) -> Iterable:
        """Read Hissp forms from code string."""
        res: Iterable[object] = self.parse(Lexer(code, self.filename))
        self.reinit()
        if self.verbose:
            res = list(res)
            pprint(res)
        return res

    def compile(self, code: str) -> str:
        """Read Lissp code and pass it on to the Hissp compiler."""
        hissp = self.reads(code)
        return self.compiler.compile(hissp)

    def gensym(self, form: str):
        """Generate a symbol unique to the current template."""
        try:
            return f"_{munge(form)}xAUTO{self.gensym_stack[-1]}_"
        except IndexError:
            raise SyntaxError("Gensym outside of template.", self.position()) from None

    @contextmanager
    def gensym_context(self):
        """Start a new gensym context for the current template."""
        self.gensym_stack.append(gensym_counter())
        try:
            yield
        finally:
            self.gensym_stack.pop()

    @contextmanager
    def unquote_context(self):
        """Start a new unquote context for the current template."""
        try:
            gensym_number = self.gensym_stack.pop()
        except IndexError:
            raise SyntaxError("Unquote outside of template.", self.position()) from None
        try:
            yield
        finally:
            self.gensym_stack.append(gensym_number)


def is_string(form):
    """
    Determines if form could have been read from a Lissp string literal.

    It's not enough to check if the form has a string type.
    Several token types such as control words, symbols, and Python
    injections, read in as strings. Macros may need to distinguish these
    cases.
    """
    try:
        return (
            type(form) is str
            and form.startswith("(")
            and type(ast.literal_eval(form)) is str
        )
    except:
        return False


def transpile(package: Optional[resources.Package], *modules: Union[str, PurePath]):
    """
    Compiles the named modules to Python files.
    If the package is None or "", it uses the current working directory without using a package.
    Lissp files must know their package at compile time to resolve imports correctly.
    """
    # TODO: allow pathname without + ".lissp"?
    if package:
        for module in modules:
            transpile_module(package, module + ".lissp")
    else:
        for module in modules:
            with open(module + ".lissp") as f:
                code = f.read()
            out = module + ".py"
            _write_py(out, module, code)


def transpile_module(
    package: resources.Package,
    resource: Union[str, PurePath],
    out: Union[None, str, bytes, Path] = None,
):
    """Transpile a single submodule in a package."""
    code = resources.read_text(package, resource)
    path: Path
    with resources.path(package, resource) as path:
        out = out or path.with_suffix(".py")
        if isinstance(package, ModuleType):
            package = package.__package__
        if isinstance(package, os.PathLike):
            resource = resource.stem
        _write_py(out, f"{package}.{resource.split('.')[0]}", code)


def _write_py(out, qualname, code):
    with open(out, "w") as f:
        print(f"compiling {qualname} as", out, file=sys.stderr)
        if code.startswith("#!"):  # ignore shebang line
            _, _, code = code.partition("\n")
        f.write(Lissp(qualname, evaluate=True, filename=str(out)).compile(code))


def main():
    """Calls `transpile` with arguments from `sys.argv`."""
    transpile(*sys.argv[1:])


if __name__ == "__main__":
    # TODO: test CLI
    main()



# Copyright 2019, 2020 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0
"""
Lissp's symbol munger.

Encodes Lissp symbols with special characters into valid,
human-readable (if ugly) Python identifiers,
using NFKC normalization and *x-codes*.

E.g. ``*FOO-BAR*`` becomes ``xSTAR_FOOxH_BARxSTAR_``.

X-codes are written in upper case and wrapped in an ``x`` and ``_``.
This format was chosen because it contains an underscore
and both lower-case and upper-case letters,
which makes it distinct from
`standard Python naming conventions`__:
``lower_case_with_underscores``,
``UPPER_CASE_WITH_UNDERSCORES``,
and ``CapWords``,
which makes the x-encoding (but not the normalization)
reversible in the usual cases,
and also cannot introduce a leading underscore,
which can have special meaning in Python.

__ https://www.python.org/dev/peps/pep-0008/#naming-conventions

Characters can be encoded in one of three ways:
Short names, Unicode names, and ordinals.

The `demunge` function will accept any of these encodings,
while the `munge` function will prioritize short names,
then fall back to Unicode names, then fall back to ordinals.

Short names are given in the `TO_NAME` table in this module.

Any spaces in the Unicode names are replaced with an ``x`` and
any hyphens are replaced with an ``h``.
(Unicode names are in all caps and these substitutions are lower-case.)

Ordinals are given in base 10.
"""

import re
import unicodedata
from contextlib import suppress
from typing import Dict, Hashable, Mapping, Match, TypeVar


def munge(s: str) -> str:
    """
    Lissp's symbol munger.

    Encodes Lissp symbols with special characters into valid,
    human-readable (if ugly) Python identifiers,
    using NFKC normalization and *x-codes*.

    Inputs that begin with ``:`` are assumed to be control words
    and returned unmodified.
    Full stops are handled separately, as those are meaningful to Hissp.
    """
    if s.startswith(":"):
        return s  # control word
    # Always normalize identifiers:
    # >>> 𝐀 = 'MATHEMATICAL BOLD CAPITAL A'
    # >>> 'A' in globals()
    # True
    s = unicodedata.normalize("NFKC", s)
    if s.isidentifier():
        return s  # Nothing to munge.
    return ".".join(_munge_part(part) for part in s.split("."))


def _munge_part(part):
    if part:
        part = "".join(map(x_encode, part))
        if not part.isidentifier():
            part = force_x_encode(part[0]) + part[1:]
            assert part.isidentifier(), f"{part!r} is not identifier"
    return part


TO_NAME = {
    # ASCII control characters don't munge to names.
    "!": "xBANG_",
    '"': "x2QUOTE_",
    "#": "xHASH_",
    "$": "xDOLR_",
    "%": "xPCENT_",
    "&": "xET_",
    "'": "x1QUOTE_",
    "(": "xPAREN_",
    ")": "xTHESES_",
    "*": "xSTAR_",
    "+": "xPLUS_",
    # xCOMMA_ is fine.
    "-": "xH_",  # Hyphen-minus
    # Full stop reserved for imports and attributes.
    "/": "xSLASH_",
    # Digits only munge if first character.
    ";": "xSCOLON_",
    "<": "xLT_",  # Less Than or LefT.
    "=": "xEQ_",
    ">": "xGT_",  # Greater Than or riGhT.
    "?": "xQUERY_",
    "@": "xAT_",
    # Capital letters are always valid in Python identifiers.
    "[": "xSQUARE_",
    "\\": "xBSLASH_",
    "]": "xBRACKETS_",
    "^": "xCARET_",
    # Underscore is valid in Python identifiers.
    "`": "xGRAVE_",
    # Small letters are also always valid.
    "{": "xCURLY_",
    "|": "xBAR_",
    "}": "xBRACES_",
    # xTILDE_ is fine.
}
"""Shorter names for X-encoding."""

X_NAME = {ord(k): ord(v) for k, v in {" ": "x", "-": "h"}.items()}


def x_encode(c: str) -> str:
    """
    Converts a character to its short x-encoding,
    unless it's already valid in a Python identifier.
    """
    if ("x" + c).isidentifier():
        return c
    return force_x_encode(c)


def force_x_encode(c: str) -> str:
    """
    Converts a character to its x-encoding,
    even if it's valid in a Python identifier.
    """
    with suppress(LookupError):
        return TO_NAME[c]
    with suppress(ValueError):
        return f"x{unicodedata.name(c).translate(X_NAME)}_"
    return f"x{ord(c)}_"


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def _reversed_1to1(mapping: Mapping[K, V]) -> Dict[V, K]:
    result = {v: k for k, v in mapping.items()}
    assert len(mapping) == len(result)
    return result


LOOKUP_NAME = _reversed_1to1(TO_NAME)
"""The inverse of `TO_NAME`."""

UN_X_NAME = _reversed_1to1(X_NAME)


def _x_decode(match: Match[str]) -> str:
    with suppress(KeyError):
        return LOOKUP_NAME[match.group()]
    with suppress(KeyError):
        return unicodedata.lookup(match.group(1).translate(UN_X_NAME))
    with suppress(ValueError):
        return chr(int(match.group(1)))
    return match.group()


def demunge(s: str) -> str:
    """The inverse of `munge`. Decodes any x-codes into characters."""
    return re.sub("x([0-9A-Zhx]+?)_", _x_decode, s)



(" Hissp's basic macros.\n"
 '\n'
 'These are automatically made available as unqualified macros in the\n'
 'Lissp REPL. To use them in a Hissp module, either use the\n'
 'fully-qualified names, or start with the `prelude` macro. You can\n'
 'abbreviate qualifiers with the `alias` macro:\n'
 '\n'
 '.. code-block:: Lissp\n'
 '\n'
 '  (hissp.basic.._macro_.alias b/ hissp.basic.._macro_.)\n'
 '  ;; Now the same as (hissp.basic.._macro_.define foo 2).\n'
 '  (b/#define foo 2)\n'
 '\n'
 'The basic macros are deliberately restricted in design.\n'
 '\n'
 'They have NO DEPENDENCIES in their expansions; they use only the\n'
 'standard library with no extra helper functions. This means that all\n'
 'helper code must be inlined, resulting in larger expansions than might\n'
 'otherwise be necessary. But because macros expand before runtime, the\n'
 'compiled code does not require Hissp to be installed to work, assuming,\n'
 'of course, that there are no other runtime imports of the library code.\n'
 '\n'
 'These may suffice for small projects, but projects with access to better\n'
 'alternatives need not use the basic macros at all.\n'
 '\n'
 'They also have no prerequisite initialization, beyond what is available\n'
 'in a standard Python module. For example, a ``_macro_`` namespace need\n'
 "not be available for ``defmacro``. It's smart enough to check for the\n"
 'presence of ``_macro_`` in its expansion context, and inline the\n'
 'initialization code when required.\n'
 '\n'
 'With the exception of ``prelude`` (which is only meant for the top\n'
 'level), they also eschew any expansions directly to Python code, relying\n'
 'only on the built-in special forms ``quote`` and ``lambda``, which makes\n'
 'their expansions compatible with advanced rewriting macros that process\n'
 'the Hissp expansions of other macros.\n'
 '\n'
 'To help keep macros and their expansions manageable in complexity,\n'
 'these basic macros lack some of the extra features their equivalents\n'
 'have in Python or in other Lisps.\n')

__import__('operator').setitem(
  globals(),
  '_macro_',
  __import__('types').ModuleType(
    '_macro_'))

setattr(
  _macro_,
  'defmacro',
  (lambda name,parameters,docstring,*body:
    (lambda *xAUTO0_:xAUTO0_)(
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (lambda *xAUTO0_:xAUTO0_)(
          ':',
          '_GxAUTO1_',
          (lambda *xAUTO0_:xAUTO0_)(
            'lambda',
            parameters,
            *body)),
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..setattr',
          '_GxAUTO1_',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            '__doc__'),
          docstring),
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..setattr',
          '_GxAUTO1_',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            '__qualname__'),
          (lambda *xAUTO0_:xAUTO0_)(
            '.join',
            "('.')",
            (lambda *xAUTO0_:xAUTO0_)(
              'quote',
              (lambda *xAUTO0_:xAUTO0_)(
                '_macro_',
                name)))),
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..setattr',
          'hissp.basic.._macro_',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            name),
          '_GxAUTO1_')))))

# defmacro
(lambda _GxAUTO1_=(lambda test,then,otherwise:
  (lambda *xAUTO0_:xAUTO0_)(
    (lambda *xAUTO0_:xAUTO0_)(
      'lambda',
      (lambda *xAUTO0_:xAUTO0_)(
        'test',
        ':',
        ':*',
        'thenxH_else'),
      (lambda *xAUTO0_:xAUTO0_)(
        (lambda *xAUTO0_:xAUTO0_)(
          'operator..getitem',
          'thenxH_else',
          (lambda *xAUTO0_:xAUTO0_)(
            'operator..not_',
            'test')))),
    test,
    (lambda *xAUTO0_:xAUTO0_)(
      'lambda',
      ':',
      then),
    (lambda *xAUTO0_:xAUTO0_)(
      'lambda',
      ':',
      otherwise))):(
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__doc__',
    ('``if-else`` Basic ternary branching construct.\n'
     '\n'
     "  Like Python's conditional expressions, the 'else' clause is required.\n"
     '  ')),
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'ifxH_else'))),
  __import__('builtins').setattr(
    __import__('builtins').globals()['_macro_'],
    'ifxH_else',
    _GxAUTO1_))[-1])()

# defmacro
(lambda _GxAUTO1_=(lambda *body:
  (lambda *xAUTO0_:xAUTO0_)(
    (lambda *xAUTO0_:xAUTO0_)(
      'lambda',
      ':',
      *body))):(
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__doc__',
    ('Evaluates each form in sequence for side effects.\n'
     '\n'
     '  Evaluates to the same value as its last form (or ``()`` if empty).\n'
     '  ')),
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'progn'))),
  __import__('builtins').setattr(
    __import__('builtins').globals()['_macro_'],
    'progn',
    _GxAUTO1_))[-1])()

# defmacro
(lambda _GxAUTO1_=(lambda condition,*body:
  (lambda *xAUTO0_:xAUTO0_)(
    'hissp.basic.._macro_.ifxH_else',
    condition,
    (lambda *xAUTO0_:xAUTO0_)(
      'hissp.basic.._macro_.progn',
      *body),
    ())):(
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__doc__',
    ('When the condition is true,\n'
     '  evaluates each form in sequence for side effects.\n'
     '  Evaluates to the same value as its last form.\n'
     '  Otherwise, skips them and returns ``()``.\n'
     '  ')),
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'when'))),
  __import__('builtins').setattr(
    __import__('builtins').globals()['_macro_'],
    'when',
    _GxAUTO1_))[-1])()

# defmacro
(lambda _GxAUTO1_=(lambda condition,*body:
  (lambda *xAUTO0_:xAUTO0_)(
    'hissp.basic.._macro_.ifxH_else',
    condition,
    (),
    (lambda *xAUTO0_:xAUTO0_)(
      'hissp.basic.._macro_.progn',
      *body))):(
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__doc__',
    ('Unless the condition is true,\n'
     '  evaluates each form in sequence for side effects.\n'
     '  Evaluates to the same value as its last form.\n'
     '  Otherwise, skips them and returns ``()``.\n'
     '  ')),
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'unless'))),
  __import__('builtins').setattr(
    __import__('builtins').globals()['_macro_'],
    'unless',
    _GxAUTO1_))[-1])()

# defmacro
(lambda _GxAUTO1_=(lambda pairs,*body:
  (lambda *xAUTO0_:xAUTO0_)(
    (lambda *xAUTO0_:xAUTO0_)(
      'lambda',
      (lambda *xAUTO0_:xAUTO0_)(
        ':',
        *pairs),
      *body))):(
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__doc__',
    ('Creates locals. Pairs are implied. Locals are not in scope until the body.')),
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'let'))),
  __import__('builtins').setattr(
    __import__('builtins').globals()['_macro_'],
    'let',
    _GxAUTO1_))[-1])()

# defmacro
(lambda _GxAUTO1_=(lambda name,parameters,docstring=(),*body:
  # let
  (lambda xDOLR_fn='_fnxAUTO7_':
    # let
    (lambda fn=(lambda *xAUTO0_:xAUTO0_)(
      'lambda',
      parameters,
      docstring,
      *body),ns=# unless
    # hissp.basic.._macro_.ifxH_else
    (lambda test,*thenxH_else:
      __import__('operator').getitem(
        thenxH_else,
        __import__('operator').not_(
          test))())(
      __import__('operator').contains(
        __import__('hissp.compiler',fromlist='?').NS.get(),
        '_macro_'),
      (lambda :()),
      (lambda :
        # hissp.basic.._macro_.progn
        (lambda :
          (lambda *xAUTO0_:xAUTO0_)(
            (lambda *xAUTO0_:xAUTO0_)(
              'operator..setitem',
              (lambda *xAUTO0_:xAUTO0_)(
                'builtins..globals'),
              (lambda *xAUTO0_:xAUTO0_)(
                'quote',
                '_macro_'),
              (lambda *xAUTO0_:xAUTO0_)(
                'types..ModuleType',
                (lambda *xAUTO0_:xAUTO0_)(
                  'quote',
                  '_macro_')))))())),dc=# when
    # hissp.basic.._macro_.ifxH_else
    (lambda test,*thenxH_else:
      __import__('operator').getitem(
        thenxH_else,
        __import__('operator').not_(
          test))())(
      __import__('hissp.reader',fromlist='?').is_string(
        docstring),
      (lambda :
        # hissp.basic.._macro_.progn
        (lambda :
          (lambda *xAUTO0_:xAUTO0_)(
            (lambda *xAUTO0_:xAUTO0_)(
              'builtins..setattr',
              xDOLR_fn,
              (lambda *xAUTO0_:xAUTO0_)(
                'quote',
                '__doc__'),
              docstring)))()),
      (lambda :())),qn=(lambda *xAUTO0_:xAUTO0_)(
      'builtins..setattr',
      xDOLR_fn,
      (lambda *xAUTO0_:xAUTO0_)(
        'quote',
        '__qualname__'),
      (lambda *xAUTO0_:xAUTO0_)(
        '.join',
        "('.')",
        (lambda *xAUTO0_:xAUTO0_)(
          'quote',
          (lambda *xAUTO0_:xAUTO0_)(
            '_macro_',
            name)))):
      (lambda *xAUTO0_:xAUTO0_)(
        'hissp.basic.._macro_.let',
        (lambda *xAUTO0_:xAUTO0_)(
          xDOLR_fn,
          fn),
        *ns,
        *dc,
        qn,
        (lambda *xAUTO0_:xAUTO0_)(
          'builtins..setattr',
          '_macro_',
          (lambda *xAUTO0_:xAUTO0_)(
            'quote',
            name),
          xDOLR_fn)))())()):(
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__doc__',
    ('Creates a new macro for the current module.\n'
     '\n'
     "  If there's no ``_macro_``, creates one (using `types.ModuleType`).\n"
     "  If there's a docstring, stores it as the new lambda's ``__doc__``.\n"
     "  Adds the ``_macro_`` prefix to the lambda's ``__qualname__``.\n"
     '  Saves the lambda in ``_macro_`` using the given attribute name.\n'
     '  ')),
  __import__('builtins').setattr(
    _GxAUTO1_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'defmacro'))),
  __import__('builtins').setattr(
    __import__('builtins').globals()['_macro_'],
    'defmacro',
    _GxAUTO1_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda name,value:(
  ('Assigns a global in the current module.'),
  (lambda *xAUTO0_:xAUTO0_)(
    'operator..setitem',
    (lambda *xAUTO0_:xAUTO0_)(
      'builtins..globals'),
    (lambda *xAUTO0_:xAUTO0_)(
      'quote',
      name),
    value))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Assigns a global in the current module.')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'define'))),
  __import__('builtins').setattr(
    _macro_,
    'define',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda name,bases,*body:(
  ('Defines a type (class) in the current module.\n'
   '\n'
   '  Key-value pairs are implied in the body.\n'
   '  '),
  (lambda *xAUTO0_:xAUTO0_)(
    'hissp.basic.._macro_.define',
    name,
    (lambda *xAUTO0_:xAUTO0_)(
      'builtins..type',
      (lambda *xAUTO0_:xAUTO0_)(
        'quote',
        name),
      (lambda *xAUTO0_:xAUTO0_)(
        (lambda *xAUTO0_:xAUTO0_)(
          'lambda',
          (lambda *xAUTO0_:xAUTO0_)(
            ':',
            ':*',
            'xAUTO0_'),
          'xAUTO0_'),
        *bases),
      (lambda *xAUTO0_:xAUTO0_)(
        'builtins..dict',
        ':',
        *body))))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Defines a type (class) in the current module.\n'
     '\n'
     '  Key-value pairs are implied in the body.\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'deftype'))),
  __import__('builtins').setattr(
    _macro_,
    'deftype',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda *args:
  (lambda *xAUTO0_:xAUTO0_)(
    'builtins..print',
    ("('Nothingness above abstraction\\n'\n"
     " '  but implementation is / the best name.\\n'\n"
     " 'Terseness may make one too many / get used to them\\n'\n"
     " '  else overflow your brain.\\n'\n"
     " 'No substitute for understanding\\n'\n"
     " 'Code;     the liability\\n'\n"
     " 'as asset; the   ability.\\n'\n"
     " 'The biggest chunks / are hard to swallow\\n'\n"
     " '  as simple as possible / no more.\\n'\n"
     " 'Source was made / for the human\\n'\n"
     " '  object / the machine.\\n'\n"
     " 'Are you lazy enough to bear / the sincerest form / of other ways of "
     "being?\\n'\n"
     " '*having* decent standards / is more important / than exactly what they "
     "are\\n'\n"
     " 'Perfection / is expensive\\n'\n"
     " '  magic / highly prized\\n'\n"
     ' "  pay for when / it\'s Worth It\\n"\n'
     " '  a quarter is advised\\n'\n"
     " 'Readability / is mainly / laid out on the page.\\n'\n"
     " 'Golfing / makes good practice / best practice it betrays.\\n'\n"
     " 'Castles built / in the air / where they do belong\\n'\n"
     " '  Elegance / then exception\\n'\n"
     " '  Form / before detail\\n'\n"
     " '  whence under them,\\n'\n"
     " 'Foundations appear.\\n'\n"
     " 'Make the right way obvious,\\n'\n"
     " 'meditate on this.\\n'\n"
     " '  --Za Zen of Hissp\\n')"))):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'import'))),
  __import__('builtins').setattr(
    _macro_,
    'import',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda target,*args:(
  ('Attaches the named variables to the target as attributes.\n'
   '\n'
   '  Positional arguments must be unqualified identifiers,\n'
   '  and use that as the attribute name.\n'
   '  Names after the ``:`` are attribute-value pairs.\n'
   '  Returns the target.\n'
   '  '),
  # let
  (lambda iargs=iter(
    args),xDOLR_target='_targetxAUTO16_':
    # let
    (lambda args=__import__('itertools').takewhile(
      (lambda a:
        __import__('operator').ne(
          a,
          ':')),
      iargs):
      (lambda *xAUTO0_:xAUTO0_)(
        'hissp.basic.._macro_.let',
        (lambda *xAUTO0_:xAUTO0_)(
          xDOLR_target,
          target),
        *map(
          (lambda arg:
            (lambda *xAUTO0_:xAUTO0_)(
              'builtins..setattr',
              xDOLR_target,
              (lambda *xAUTO0_:xAUTO0_)(
                'quote',
                arg),
              arg)),
          args),
        *map(
          (lambda kw:
            (lambda *xAUTO0_:xAUTO0_)(
              'builtins..setattr',
              xDOLR_target,
              (lambda *xAUTO0_:xAUTO0_)(
                'quote',
                kw),
              next(
                iargs))),
          iargs),
        xDOLR_target))())())[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Attaches the named variables to the target as attributes.\n'
     '\n'
     '  Positional arguments must be unqualified identifiers,\n'
     '  and use that as the attribute name.\n'
     '  Names after the ``:`` are attribute-value pairs.\n'
     '  Returns the target.\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'attach'))),
  __import__('builtins').setattr(
    _macro_,
    'attach',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda self,*calls:(
  ('Call multiple methods on one object.\n'
   '\n'
   '  Evaluates the given ``self``, then injects it as the first argument to\n'
   '  a sequence of calls. Returns the result of the last call.\n'
   '  '),
  # let
  (lambda xDOLR_self='_selfxAUTO20_':
    (lambda *xAUTO0_:xAUTO0_)(
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (lambda *xAUTO0_:xAUTO0_)(
          ':',
          xDOLR_self,
          self),
        *map(
          (lambda call:
            (lambda *xAUTO0_:xAUTO0_)(
              __import__('operator').getitem(
                call,
                (0)),
              xDOLR_self,
              *__import__('operator').getitem(
                call,
                slice(
                  (1),
                  None)))),
          calls))))())[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Call multiple methods on one object.\n'
     '\n'
     '  Evaluates the given ``self``, then injects it as the first argument to\n'
     '  a sequence of calls. Returns the result of the last call.\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'cascade'))),
  __import__('builtins').setattr(
    _macro_,
    'cascade',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda expr,*forms:(
  ("``->`` 'Thread-first'.\n"
   '\n'
   '  Converts a pipeline to function calls by recursively threading it as\n'
   '  the first argument of the next form.\n'
   '  E.g. ``(-> x (A b) (C d e))`` is ``(C (A x b) d e)``\n'
   '  Makes chained method calls easier to read.\n'
   '  '),
  # ifxH_else
  (lambda test,*thenxH_else:
    __import__('operator').getitem(
      thenxH_else,
      __import__('operator').not_(
        test))())(
    forms,
    (lambda :
      (lambda *xAUTO0_:xAUTO0_)(
        'hissp.basic..xAUTO_.xH_xGT_',
        (lambda *xAUTO0_:xAUTO0_)(
          __import__('operator').getitem(
            __import__('operator').getitem(
              forms,
              (0)),
            (0)),
          expr,
          *__import__('operator').getitem(
            __import__('operator').getitem(
              forms,
              (0)),
            slice(
              (1),
              None))),
        *__import__('operator').getitem(
          forms,
          slice(
            (1),
            None)))),
    (lambda :expr)))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ("``->`` 'Thread-first'.\n"
     '\n'
     '  Converts a pipeline to function calls by recursively threading it as\n'
     '  the first argument of the next form.\n'
     '  E.g. ``(-> x (A b) (C d e))`` is ``(C (A x b) d e)``\n'
     '  Makes chained method calls easier to read.\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'xH_xGT_'))),
  __import__('builtins').setattr(
    _macro_,
    'xH_xGT_',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda expr,*forms:(
  ("``->>`` 'Thread-last'.\n"
   '\n'
   '  Converts a pipeline to function calls by recursively threading it as\n'
   '  the last argument of the next form.\n'
   '  E.g. ``(->> x (A b) (C d e))`` is ``(C d e (A b x))``.\n'
   '  Can replace partial application in some cases.\n'
   '  Also works inside a ``->`` pipeline.\n'
   '  E.g. ``(-> x (A a) (->> B b) (C c))`` is ``(C (B b (A x a)) c)``.\n'
   '  '),
  # ifxH_else
  (lambda test,*thenxH_else:
    __import__('operator').getitem(
      thenxH_else,
      __import__('operator').not_(
        test))())(
    forms,
    (lambda :
      (lambda *xAUTO0_:xAUTO0_)(
        'hissp.basic..xAUTO_.xH_xGT_xGT_',
        (lambda *xAUTO0_:xAUTO0_)(
          *__import__('operator').getitem(
            forms,
            (0)),
          expr),
        *__import__('operator').getitem(
          forms,
          slice(
            (1),
            None)))),
    (lambda :expr)))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ("``->>`` 'Thread-last'.\n"
     '\n'
     '  Converts a pipeline to function calls by recursively threading it as\n'
     '  the last argument of the next form.\n'
     '  E.g. ``(->> x (A b) (C d e))`` is ``(C d e (A b x))``.\n'
     '  Can replace partial application in some cases.\n'
     '  Also works inside a ``->`` pipeline.\n'
     '  E.g. ``(-> x (A a) (->> B b) (C c))`` is ``(C (B b (A x a)) c)``.\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'xH_xGT_xGT_'))),
  __import__('builtins').setattr(
    _macro_,
    'xH_xGT_xGT_',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda *pairs:(
  ('Multiple condition branching.\n'
   '\n'
   '  Pairs are implied. Default is ``()``. Use ``:else`` to change it.\n'
   '  For example::\n'
   '\n'
   '   (cond)  ; ()\n'
   "   ;; Assume some number 'x\n"
   '   (cond (operator..gt x 0) (print "positive")\n'
   '         (operator..lt x 0) (print "negative")\n'
   '         (operator..eq x 0) (print "zero")\n'
   '         :else (print "not a number"))\n'
   '  '),
  # when
  # hissp.basic.._macro_.ifxH_else
  (lambda test,*thenxH_else:
    __import__('operator').getitem(
      thenxH_else,
      __import__('operator').not_(
        test))())(
    pairs,
    (lambda :
      # hissp.basic.._macro_.progn
      (lambda :
        (lambda *xAUTO0_:xAUTO0_)(
          'hissp.basic.._macro_.ifxH_else',
          __import__('operator').getitem(
            pairs,
            (0)),
          __import__('operator').getitem(
            pairs,
            (1)),
          (lambda *xAUTO0_:xAUTO0_)(
            'hissp.basic..xAUTO_.cond',
            *__import__('operator').getitem(
              pairs,
              slice(
                (2),
                None)))))()),
    (lambda :())))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Multiple condition branching.\n'
     '\n'
     '  Pairs are implied. Default is ``()``. Use ``:else`` to change it.\n'
     '  For example::\n'
     '\n'
     '   (cond)  ; ()\n'
     "   ;; Assume some number 'x\n"
     '   (cond (operator..gt x 0) (print "positive")\n'
     '         (operator..lt x 0) (print "negative")\n'
     '         (operator..eq x 0) (print "zero")\n'
     '         :else (print "not a number"))\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'cond'))),
  __import__('builtins').setattr(
    _macro_,
    'cond',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda variable,iterable,*body:(
  ('``any-for``\n'
   '  Bind the variable and evaluate the body for each item from the\n'
   '  iterable until any result is true (and return ``True``),\n'
   '  or until the iterable is exhausted (and return ``False``).\n'
   '  '),
  (lambda *xAUTO0_:xAUTO0_)(
    'builtins..any',
    (lambda *xAUTO0_:xAUTO0_)(
      'builtins..map',
      (lambda *xAUTO0_:xAUTO0_)(
        'lambda',
        (lambda *xAUTO0_:xAUTO0_)(
          variable),
        *body),
      iterable)))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('``any-for``\n'
     '  Bind the variable and evaluate the body for each item from the\n'
     '  iterable until any result is true (and return ``True``),\n'
     '  or until the iterable is exhausted (and return ``False``).\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'anyxH_for'))),
  __import__('builtins').setattr(
    _macro_,
    'anyxH_for',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda *exprs:(
  ("``&&`` 'and'. Like Python's ``and`` operator, but for any number of arguments."),
  # cond
  # hissp.basic.._macro_.ifxH_else
  (lambda test,*thenxH_else:
    __import__('operator').getitem(
      thenxH_else,
      __import__('operator').not_(
        test))())(
    __import__('operator').not_(
      exprs),
    (lambda :True),
    (lambda :
      # hissp.basic..xAUTO_.cond
      # hissp.basic.._macro_.ifxH_else
      (lambda test,*thenxH_else:
        __import__('operator').getitem(
          thenxH_else,
          __import__('operator').not_(
            test))())(
        __import__('operator').eq(
          len(
            exprs),
          (1)),
        (lambda :
          __import__('operator').getitem(
            exprs,
            (0))),
        (lambda :
          # hissp.basic..xAUTO_.cond
          # hissp.basic.._macro_.ifxH_else
          (lambda test,*thenxH_else:
            __import__('operator').getitem(
              thenxH_else,
              __import__('operator').not_(
                test))())(
            ':else',
            (lambda :
              (lambda *xAUTO0_:xAUTO0_)(
                'hissp.basic.._macro_.let',
                (lambda *xAUTO0_:xAUTO0_)(
                  '_GxAUTO27_',
                  __import__('operator').getitem(
                    exprs,
                    (0))),
                (lambda *xAUTO0_:xAUTO0_)(
                  'hissp.basic.._macro_.ifxH_else',
                  '_GxAUTO27_',
                  (lambda *xAUTO0_:xAUTO0_)(
                    'hissp.basic..xAUTO_.xET_xET_',
                    *__import__('operator').getitem(
                      exprs,
                      slice(
                        (1),
                        None))),
                  '_GxAUTO27_'))),
            (lambda :
              # hissp.basic..xAUTO_.cond
              ())))))))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ("``&&`` 'and'. Like Python's ``and`` operator, but for any number of arguments.")),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'xET_xET_'))),
  __import__('builtins').setattr(
    _macro_,
    'xET_xET_',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda first=(),*rest:(
  ("``||`` 'or'. Like Python's ``or`` operator, but for any number of arguments."),
  # ifxH_else
  (lambda test,*thenxH_else:
    __import__('operator').getitem(
      thenxH_else,
      __import__('operator').not_(
        test))())(
    rest,
    (lambda :
      (lambda *xAUTO0_:xAUTO0_)(
        'hissp.basic.._macro_.let',
        (lambda *xAUTO0_:xAUTO0_)(
          '_firstxAUTO28_',
          first),
        (lambda *xAUTO0_:xAUTO0_)(
          'hissp.basic.._macro_.ifxH_else',
          '_firstxAUTO28_',
          '_firstxAUTO28_',
          (lambda *xAUTO0_:xAUTO0_)(
            'hissp.basic..xAUTO_.xBAR_xBAR_',
            *rest)))),
    (lambda :first)))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ("``||`` 'or'. Like Python's ``or`` operator, but for any number of arguments.")),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'xBAR_xBAR_'))),
  __import__('builtins').setattr(
    _macro_,
    'xBAR_xBAR_',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda expr1,*body:(
  ('Evaluates each expression in sequence (for side effects),\n'
   '  resulting in the value of the first.'),
  (lambda *xAUTO0_:xAUTO0_)(
    'hissp.basic.._macro_.let',
    (lambda *xAUTO0_:xAUTO0_)(
      '_value1xAUTO29_',
      expr1),
    *body,
    '_value1xAUTO29_'))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Evaluates each expression in sequence (for side effects),\n'
     '  resulting in the value of the first.')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'prog1'))),
  __import__('builtins').setattr(
    _macro_,
    'prog1',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda raw:(
  ('``b#`` bytes literal reader macro'),
  # xH_xGT_
  # hissp.basic..xAUTO_.xH_xGT_
  # hissp.basic..xAUTO_.xH_xGT_
  # hissp.basic..xAUTO_.xH_xGT_
  # hissp.basic..xAUTO_.xH_xGT_
  __import__('ast').literal_eval(
    # xH_xGT_xGT_
    # hissp.basic..xAUTO_.xH_xGT_xGT_
    ("b'{}'").format(
      raw.replace(
        ("'"),
        ("\\'")).replace(
        ('\n'),
        ('\\n')))))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('``b#`` bytes literal reader macro')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'bxHASH_'))),
  __import__('builtins').setattr(
    _macro_,
    'bxHASH_',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda :(
  ('Grants unqualified access to the basics.\n'
   '\n'
   '  Star imports from `operator` and `itertools`.\n'
   '  And adds the basic macros, but only if available,\n'
   '  so its expansion does not require Hissp to be installed.\n'
   '  (This overwrites ``_macro_`` if you already had one.)\n'
   '  '),
  (lambda *xAUTO0_:xAUTO0_)(
    'builtins..exec',
    ("('from operator import *\\n'\n"
     " 'from itertools import *\\n'\n"
     " 'try:\\n'\n"
     " '    from hissp.basic import _macro_\\n'\n"
     ' "    _macro_ = __import__(\'types\').SimpleNamespace(**vars(_macro_))\\n"\n'
     " 'except ModuleNotFoundError:\\n'\n"
     " '    pass')")))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Grants unqualified access to the basics.\n'
     '\n'
     '  Star imports from `operator` and `itertools`.\n'
     '  And adds the basic macros, but only if available,\n'
     '  so its expansion does not require Hissp to be installed.\n'
     '  (This overwrites ``_macro_`` if you already had one.)\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'prelude'))),
  __import__('builtins').setattr(
    _macro_,
    'prelude',
    _fnxAUTO7_))[-1])()

# defmacro
# hissp.basic.._macro_.let
(lambda _fnxAUTO7_=(lambda alias,module:(
  ('Defines a reader macro abbreviation of a symbol prefix\n'
   '  (such as a qualifier). For example,\n'
   '\n'
   '  .. code-block:: Lissp\n'
   '\n'
   '     (hissp.basic.._macro_.alias b/ hissp.basic.._macro_.)\n'
   '     ;; Now the same as (hissp.basic.._macro_.define foo 2).\n'
   '     (b/#define foo 2)\n'
   '  '),
  (lambda *xAUTO0_:xAUTO0_)(
    'hissp.basic.._macro_.defmacro',
    ('{}{}').format(
      alias,
      'xHASH_'),
    (lambda *xAUTO0_:xAUTO0_)(
      '_GxAUTO31_'),
    (lambda *xAUTO0_:xAUTO0_)(
      'quote',
      ('Aliases {} as {}#').format(
        module,
        alias)),
    (lambda *xAUTO0_:xAUTO0_)(
      '.format',
      "('{}{}')",
      (lambda *xAUTO0_:xAUTO0_)(
        'quote',
        module),
      '_GxAUTO31_')))[-1]):(
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__doc__',
    ('Defines a reader macro abbreviation of a symbol prefix\n'
     '  (such as a qualifier). For example,\n'
     '\n'
     '  .. code-block:: Lissp\n'
     '\n'
     '     (hissp.basic.._macro_.alias b/ hissp.basic.._macro_.)\n'
     '     ;; Now the same as (hissp.basic.._macro_.define foo 2).\n'
     '     (b/#define foo 2)\n'
     '  ')),
  __import__('builtins').setattr(
    _fnxAUTO7_,
    '__qualname__',
    ('.').join(
      ('_macro_', 'alias'))),
  __import__('builtins').setattr(
    _macro_,
    'alias',
    _fnxAUTO7_))[-1])()


# Copyright 2019, 2020 Matthew Egan Odendahl
# SPDX-License-Identifier: Apache-2.0

"""
The Hissp data-structure language compiler and associated helper functions.
"""

import ast
import builtins
import pickle
import pickletools
import re
import sys
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from functools import wraps
from itertools import chain, takewhile
from pprint import pformat
from traceback import format_exc
from types import ModuleType
from typing import Iterable, List, NewType, Tuple, TypeVar, Union
from warnings import warn

PAIR_WORDS = {":*": "*", ":**": "**", ":?": ""}
# Module Macro container
MACROS = "_macro_"
# Macro from foreign module foo.bar.._macro_.baz
MACRO = f"..{MACROS}."
RE_MACRO = re.compile(rf"(\.\.{MACROS}\.|\.\.xAUTO_\.)")

NS = ContextVar("NS", default=())
"""
Sometimes macros need the current namespace when expanding,
instead of its defining namespace.
Rather than pass in an implicit argument to all macros,
it's available here.
`readerless` uses this automatically.
"""

Sentinel = NewType("Sentinel", object)
_SENTINEL = Sentinel(object())


class CompileError(SyntaxError):
    """Catch-all exception for compilation failures."""


def _trace(method):
    @wraps(method)
    def tracer(self, expr):
        try:
            return method(self, expr)
        except Exception as e:
            self.error = e
            message = (
                f"\nCompiler.{method.__name__}() {type(e).__name__}:\n {e}".replace(
                    "\n", "\n# "
                )
            )
            return f"(>   >  > >>{pformat(expr)}<< <  <   <){message}"

    return tracer


class PostCompileWarning(Warning):
    """Form compiled to Python, but execution of it failed."""


class Compiler:
    """
    The Hissp recursive-descent compiler.

    Translates the Hissp data-structure language into a functional
    subset of Python.
    """

    def __init__(self, qualname="__main__", ns=None, evaluate=True):
        self.qualname = qualname
        self.ns = self.new_ns(qualname) if ns is None else ns
        self.evaluate = evaluate
        self.error = False
        self.abort = False

    @staticmethod
    def new_ns(name, doc=None, package=None):
        mod = ModuleType(name, doc)
        mod.__annotations__ = {}
        mod.__package__ = package
        mod.__builtins__ = builtins
        return vars(mod)

    def compile(self, forms: Iterable) -> str:
        """
        Compile multiple forms, and execute them them if evaluate mode enabled.
        """
        result: List[str] = []
        for form in forms:
            form = self.form(form)
            if self.error:
                e = self.error
                self.error = False
                raise CompileError("\n" + form) from e
            result.extend(self.eval(form))
            if self.abort:
                print("\n\n".join(result), file=sys.stderr)
                sys.exit(1)
        return "\n\n".join(result)

    @_trace
    def form(self, form) -> str:
        """
        Compile Hissp form to the equivalent Python code as a string.
        `tuple` and `str` have special evaluation rules,
        otherwise it's an `atom` that represents itself.
        """
        if type(form) is tuple and form:
            return self.tuple(form)
        if type(form) is str and not form.startswith(":"):
            return self.str(form)
        return self.atom(form)

    @_trace
    def tuple(self, form: Tuple) -> str:
        """Compile `call`, `macro`, or `special` forms."""
        head, *tail = form
        if type(head) is str:
            return self.special(form)
        return self.call(form)

    @_trace
    def special(self, form: Tuple) -> str:
        """Try to compile as special form, else `invocation`."""
        if form[0] == "quote":
            return self.atom(*form[1:])
        if form[0] == "lambda":
            return self.function(form)
        return self.invocation(form)

    @_trace
    def function(self, form: Tuple) -> str:
        r"""
        Compile the anonymous function special form.

        (lambda (<parameters>)
          <body>)

        The parameters tuple is divided into (<single> : <paired>)

        Parameter types are the same as Python's.
        For example,

        >>> readerless(
        ... ('lambda', ('a',':/','b',
        ...         ':', 'e',1, 'f',2,
        ...         ':*','args', 'h',4, 'i',':?', 'j',1,
        ...         ':**','kwargs',),
        ...   42,),
        ... )
        '(lambda a,/,b,e=(1),f=(2),*args,h=(4),i,j=(1),**kwargs:(42))'

        The special control words :* and :** designate the remainder of the
        positional and keyword parameters, respectively.
        Note this body has an implicit PROGN:

        >>> print(readerless(
        ... ('lambda', (':',':*','args',':**','kwargs',),
        ...   ('print','args',),
        ...   ('print','kwargs',),),
        ... ))
        (lambda *args,**kwargs:(
          print(
            args),
          print(
            kwargs))[-1])

        You can omit the right of a pair with ``:?``
        (except the final ``**kwargs``).
        Also note that the body can be empty.

        >>> readerless(
        ... ('lambda', (':','a',1, ':/',':?', ':*',':?', 'b',':?', 'c',2,),),
        ... )
        '(lambda a=(1),/,*,b,c=(2):())'

        The ':' may be omitted if there are no paired parameters.

        >>> readerless(('lambda', ('a','b','c',':',),),)
        '(lambda a,b,c:())'
        >>> readerless(('lambda', ('a','b','c',),),)
        '(lambda a,b,c:())'
        >>> readerless(('lambda', (':',),),)
        '(lambda :())'
        >>> readerless(('lambda', (),),)
        '(lambda :())'

        ``:`` is required if there are any paired parameters, even if
        there are no single parameters:

        >>> readerless(('lambda', (':',':**','kwargs',),),)
        '(lambda **kwargs:())'

        """
        fn, parameters, *body = form
        assert fn == "lambda"
        return f"(lambda {','.join(self.parameters(parameters))}:{self.body(body)})"

    @_trace
    def parameters(self, parameters: Iterable) -> Iterable[str]:
        """Process parameters to compile `function`."""
        parameters = iter(parameters)
        yield from (
            {":/": "/", ":*": "*"}.get(a, a)
            for a in takewhile(lambda a: a != ":", parameters)
        )
        for k, v in _pairs(parameters):
            if k == ":*":
                yield "*" if v == ":?" else f"*{v}"
            elif k == ":/":
                yield "/"
            elif k == ":**":
                yield f"**{v}"
            elif v == ":?":
                yield k
            else:
                yield f"{k}={self.form(v)}"

    @_trace
    def body(self, body: list) -> str:
        """Compile body of `function`."""
        if len(body) > 1:
            return f"({_join_args(*map(self.form, body))})[-1]"
        if not body:
            return "()"
        result = self.form(body[0])
        return ("\n" * ("\n" in result) + result).replace("\n", "\n  ")

    @_trace
    def invocation(self, form: Tuple) -> str:
        """Try to compile as `macro`, else normal `call`."""
        if (result := self.macro(form)) is not _SENTINEL:
            return f"# {form[0]}\n{result}"
        form = form[0].replace("..xAUTO_.", "..", 1), *form[1:]
        return self.call(form)

    @_trace
    def macro(self, form: Tuple) -> Union[str, Sentinel]:
        """Macroexpand and start over with `form`, if it's a macro."""
        head, *tail = form
        if (macro := self._get_macro(head)) is not None:
            with self.macro_context():
                return self.form(macro(*tail))
        return _SENTINEL

    def _get_macro(self, head):
        parts = RE_MACRO.split(head, 1)
        head = head.replace("..xAUTO_.", MACRO, 1)
        if len(parts) > 1:
            return self._qualified_macro(head, parts)
        return self._unqualified_macro(head)

    def _qualified_macro(self, head, parts):
        try:
            if parts[0] == self.qualname:  # Internal?
                return vars(self.ns[MACROS])[parts[2]]
            return eval(self.str(head))
        except (KeyError, AttributeError):
            if parts[1] != "..xAUTO_.":
                raise

    def _unqualified_macro(self, head):
        try:
            return vars(self.ns[MACROS])[head]
        except KeyError:
            pass

    @contextmanager
    def macro_context(self):
        """Sets `NS` during macroexpansions."""
        token = NS.set(self.ns)
        try:
            yield
        finally:
            NS.reset(token)

    @_trace
    def call(self, form: Iterable) -> str:
        r"""
        Compile call form.

        Any tuple that is not quoted, ``()``, or a `special` form or
        `macro` is a runtime call.

        Like Python, it has three parts.
        (<callable> <args> : <kwargs>)
        For example:

        >>> print(readerless(
        ... ('print',1,2,3,
        ...          ':','sep',('quote',":",), 'end',('quote',"\n\n",),)
        ... ))
        print(
          (1),
          (2),
          (3),
          sep=':',
          end='\n\n')

        Either <args> or <kwargs> may be empty:

        >>> readerless(('foo',':',),)
        'foo()'
        >>> print(readerless(('foo','bar',':',),))
        foo(
          bar)
        >>> print(readerless(('foo',':','bar','baz',),))
        foo(
          bar=baz)

        The : is optional if the <kwargs> part is empty:

        >>> readerless(('foo',),)
        'foo()'
        >>> print(readerless(('foo','bar',),),)
        foo(
          bar)

        The <kwargs> part has implicit pairs; there must be an even number.

        Use the control words ``:*`` and ``:**`` for iterable and
        mapping unpacking:

        >>> print(readerless(
        ... ('print',':',':*',[1,2], 'a',3, ':*',[4], ':**',{'sep':':','end':'\n\n'},),
        ... ))
        print(
          *[1, 2],
          a=(3),
          *[4],
          **{'sep': ':', 'end': '\n\n'})

        Unlike other control words, these can be repeated,
        but (as in Python) a '*' is not allowed to follow '**'.

        Method calls are similar to function calls.
        (.<method name> <object> <args> : <kwargs>)
        Like Clojure, a method on the first object is assumed if the
        function name starts with a dot:

        >>> readerless(('.conjugate', 1j,),)
        '(1j).conjugate()'
        >>> eval(_)
        -1j
        >>> readerless(('.decode', b'\xfffoo', ':', 'errors',('quote','ignore',),),)
        "b'\\xfffoo'.decode(\n  errors='ignore')"
        >>> eval(_)
        'foo'

        """
        form = iter(form)
        head = next(form)
        args = chain(
            map(self.form, takewhile(lambda a: a != ":", form)),
            (f"{(PAIR_WORDS.get(k, k+'='))}{self.form(v)}" for k, v in _pairs(form)),
        )
        if type(head) is str and head.startswith("."):
            return "{}.{}({})".format(next(args), head[1:], _join_args(*args))
        return "{}({})".format(self.form(head), _join_args(*args))

    @_trace
    def str(self, code: str) -> str:
        """Compile code strings.
        Expands qualified identifiers and module literals into imports.
        Otherwise, injects as raw Python directly into the output.
        """
        if re.search(r"^\.\.|[ ()]", code):  # Ellipsis? Python injection?
            return code
        if ".." in code:
            return self.qualified_identifier(code)
        elif code.endswith("."):
            return self.module_identifier(code)
        return code

    def qualified_identifier(self, code):
        """Compile qualified identifier into import and attribute."""
        parts = code.split("..", 1)
        if parts[0] == self.qualname:  # This module. No import required.
            chain = parts[1].split(".", 1)
            # Avoid local shadowing.
            chain[0] = f"__import__('builtins').globals()[{self.atom(chain[0])}]"
            return ".".join(chain)
        return "__import__({0!r}{fromlist}).{1}".format(
            parts[0], parts[1], fromlist=",fromlist='?'" if "." in parts[0] else ""
        )

    def module_identifier(self, code):
        """Compile module identifier to import."""
        module = code[:-1]
        return f"""__import__({module !r}{",fromlist='?'" if "." in module else ""})"""

    @_trace
    def atom(self, form) -> str:
        r"""
        Compile forms that evaluate to themselves.

        Emits a literal if possible, otherwise falls back to `pickle`:

        >>> readerless(-4.2j)
        '((-0-4.2j))'
        >>> print(readerless(float('nan')))
        __import__('pickle').loads(  # nan
            b'Fnan\n.'
        )
        >>> readerless([{'foo':2},(),1j,2.0,{3}])
        "[{'foo': 2}, (), 1j, 2.0, {3}]"
        >>> spam = []
        >>> spam.append(spam)
        >>> print(readerless(spam))
        __import__('pickle').loads(  # [[...]]
            b'(lp0\ng0\na.'
        )

        """
        if form is Ellipsis:
            return "..."

        case = type(form)
        if case in {int, float, complex}:  # Number literals may need (). E.g. (1).real
            literal = f"({form!r})"
        elif case in {dict, list, set, tuple, str, bytes}:  # Pretty print collections.
            literal = pformat(form, sort_dicts=False)
        else:
            literal = repr(form)

        with suppress(ValueError, SyntaxError):
            if ast.literal_eval(literal) == form:
                return literal
        # literal failed to round trip. Fall back to pickle.
        return self.pickle(form)

    @_trace
    def pickle(self, form) -> str:
        """Compile to `pickle.loads`. The final fallback for `atom`."""
        try:  # Try the more human-readable and backwards-compatible text protocol first.
            dumps = pickle.dumps(form, 0)
        except pickle.PicklingError:  # Fall back to the highest binary protocol if that didn't work.
            dumps = pickle.dumps(form, pickle.HIGHEST_PROTOCOL)
        dumps = pickletools.optimize(dumps)
        r = repr(form).replace("\n", "\n  # ")
        nl = "\n" if "\n" in r else ""
        return f"__import__('pickle').loads({nl}  # {r}\n    {dumps!r}\n)"

    def eval(self, form: str) -> Tuple[str, ...]:
        """Execute compiled form, but only if evaluate mode is enabled."""
        try:
            if self.evaluate:
                exec(compile(form, "<Hissp>", "exec"), self.ns)
        except Exception as e:
            exc = format_exc()
            if self.ns.get("__name__") == "__main__":
                self.abort = True
            else:
                warn(
                    f"\n {e} when evaluating form:\n{form}\n\n{exc}", PostCompileWarning
                )
            return form, "# " + exc.replace("\n", "\n# ")
        return (form,)


def _join_args(*args):
    return (("\n" if args else "") + ",\n".join(args)).replace("\n", "\n  ")


T = TypeVar("T")


def _pairs(it: Iterable[T]) -> Iterable[Tuple[T, T]]:
    it = iter(it)
    for k in it:
        try:
            yield k, next(it)
        except StopIteration:
            raise CompileError("Incomplete pair.") from None


def readerless(form, ns=None):
    """Compile a Hissp form to Python without evaluating it.
    Uses the current `NS` for context, unless an alternative is provided.
    (Creates a temporary namespace if neither is available.)
    Returns the Python as a string.
    """
    ns = ns or NS.get() or {"__name__": "__main__"}
    return Compiler(evaluate=False, ns=ns).compile([form])
